"""Tests for egta online scheduler"""
import asyncio
import random

import numpy as np
import pytest
from egtaonline import api
from egtaonline import mockserver
from gameanalysis import rsgame

from egta import countsched
from egta import eosched


# TODO general setup may be done with a fixture in python 3.6

@pytest.fixture(name='game')
def fix_game():
    """Fixture to create a standard game"""
    return rsgame.empty([4, 4], [11, 11])


@pytest.mark.asyncio
async def test_basic_profile(game):
    """Test scheduling a standard profile"""
    async with mockserver.server() as server, api.api() as egta:
        sim = await egta.get_simulator(server.create_simulator(
            'sim', '1', delay_dist=lambda: random.random() / 10))
        strats = dict(zip(game.role_names, game.strat_names))
        symgrps = list(zip(game.role_names, game.num_role_players,
                           game.strat_names))
        await sim.add_strategies(strats)
        egame = await egta.get_canon_game(sim['id'], symgrps)

        profs = game.random_profiles(20)

        # Schedule all new profiles and verify it works
        async with eosched.eosched(
            game, egta, egame['id'], 0.1, 1, 10, 0, 0) as sched:
            assert str(sched) == str(egame['id'])
            assert game == rsgame.empty_copy(sched)
            awaited = await asyncio.gather(*[
                sched.sample_payoffs(p) for p in profs])
        pays = np.stack(awaited)
        assert np.allclose(pays[profs == 0], 0)

        # Schedule old profiles and verify it still works
        async with eosched.eosched(
            game, egta, egame['id'], 0.1, 1, 10, 0, 0) as sched:
            awaited = await asyncio.gather(*[
                sched.sample_payoffs(p) for p in profs])
        pays = np.stack(awaited)
        assert np.allclose(pays[profs == 0], 0)

        # Schedule two at a time, in two batches
        async with eosched.eosched(
            game, egta, egame['id'], 0.1, 2, 10, 0, 0) as base_sched:
            sched = countsched.countsched(base_sched, 2)
            awaited = await asyncio.gather(*[
                sched.sample_payoffs(p) for p in profs])
        pays = np.stack(awaited)
        assert np.allclose(pays[profs == 0], 0)

        # Try again now that everything should be scheduled
        async with eosched.eosched(
            game, egta, egame['id'], 0.1, 2, 10, 0, 0) as base_sched:
            sched = countsched.countsched(base_sched, 2)
            awaited = await asyncio.gather(*[
                sched.sample_payoffs(p) for p in profs])
        pays = np.stack(awaited)
        assert np.allclose(pays[profs == 0], 0)


def _raise(ex):
    """Exception as function"""
    raise ex


@pytest.mark.asyncio
async def test_exception_in_create(game):
    """Test exception creating eo scheduler"""
    async with mockserver.server() as server, api.api() as egta:
        sim = await egta.get_simulator(server.create_simulator(  # pragma: no branch pylint: disable=line-too-long
            'sim', '1', delay_dist=lambda: random.random() / 10))
        strats = dict(zip(game.role_names, game.strat_names))
        symgrps = list(zip(game.role_names, game.num_role_players,
                           game.strat_names))
        await sim.add_strategies(strats)
        egame = await egta.get_canon_game(sim['id'], symgrps)

        server.custom_response(lambda: _raise(TimeoutError))
        with pytest.raises(TimeoutError):
            async with eosched.eosched(
                game, egta, egame['id'], 0.1, 1, 25, 0, 0):
                pass  # pragma: no cover


@pytest.mark.asyncio
async def test_exception_in_get(game):
    """Test exception in await"""
    async with mockserver.server() as server, api.api() as egta:
        sim = await egta.get_simulator(server.create_simulator(
            'sim', '1', delay_dist=lambda: random.random() / 10))
        strats = dict(zip(game.role_names, game.strat_names))
        symgrps = list(zip(game.role_names, game.num_role_players,
                           game.strat_names))
        await sim.add_strategies(strats)
        egame = await egta.get_canon_game(sim['id'], symgrps)

        profs = game.random_profiles(20)

        async with eosched.eosched(
            game, egta, egame['id'], 0.1, 1, 10, 0, 0) as sched:
            futures = asyncio.gather(*[
                sched.sample_payoffs(p) for p in profs])
            await asyncio.sleep(0.1)
            server.custom_response(lambda: _raise(TimeoutError))
            await asyncio.sleep(0.1)
            with pytest.raises(TimeoutError):
                await futures


@pytest.mark.asyncio
async def test_exception_in_schedule(game):
    """Test exception throwin in scheduler"""
    async with mockserver.server() as server, api.api() as egta:
        sim = await egta.get_simulator(server.create_simulator(  # pragma: no branch pylint: disable=line-too-long
            'sim', '1', delay_dist=lambda: random.random() / 10))
        strats = dict(zip(game.role_names, game.strat_names))
        symgrps = list(zip(game.role_names, game.num_role_players,
                           game.strat_names))
        await sim.add_strategies(strats)
        egame = await egta.get_canon_game(sim['id'], symgrps)

        prof = game.random_profile()

        async with eosched.eosched(
            game, egta, egame['id'], 0.1, 1, 25, 0, 0) as sched:
            # so that enough calls to get_requirements are made
            server.custom_response(lambda: _raise(TimeoutError))
            await asyncio.sleep(0.1)
            with pytest.raises(TimeoutError):
                await sched.sample_payoffs(prof)


@pytest.mark.asyncio
async def test_scheduler_deactivate(game):
    """Test that scheduler ends when deactivated"""
    async with mockserver.server() as server, api.api() as egta:
        sim = await egta.get_simulator(server.create_simulator(  # pragma: no branch pylint: disable=line-too-long
            'sim', '1', delay_dist=lambda: random.random() / 10))
        strats = dict(zip(game.role_names, game.strat_names))
        symgrps = list(zip(game.role_names, game.num_role_players,
                           game.strat_names))
        await sim.add_strategies(strats)
        egame = await egta.get_canon_game(sim['id'], symgrps)

        # Schedule all new profiles and verify it works
        # This first time should have to wait to schedule more
        async with eosched.eosched(
            game, egta, egame['id'], 0.1, 1, 10, 0, 0) as sched:
            # Deactivate scheduler
            for esched in await egta.get_generic_schedulers():
                await esched.deactivate()
            with pytest.raises(ValueError):
                await sched.sample_payoffs(game.random_profile())



"""Tests for count schedulers"""
import asyncio
import pytest

import numpy as np
from gameanalysis import gamegen
from gameanalysis import utils

from egta import countsched
from egta import gamesched
from egta import savesched


@pytest.mark.asyncio
async def test_basic_profile():
    """Test that basic profiles are sampled twice"""
    sgame = gamegen.samplegame([4, 3], [3, 4])
    profs = utils.axis_from_elem(np.unique(utils.axis_to_elem(
        sgame.random_profiles(20))))

    save = savesched.savesched(gamesched.samplegamesched(sgame))
    sched = countsched.countsched(save, 10)
    assert str(sched) is not None
    paylist = await asyncio.gather(*[
        sched.sample_payoffs(p) for p in profs])
    pays = np.stack(paylist)
    assert np.allclose(pays[profs == 0], 0)

    savegame = save.get_game()
    assert list(savegame.num_samples) == [10]



"""Module for tracing continuous equilibria"""
import asyncio
import functools
import itertools
import logging

import numpy as np
from scipy.sparse import csgraph
from gameanalysis import regret
from gameanalysis import rsgame
from gameanalysis import trace
from gameanalysis import utils

from egta import asyncgame
from egta import innerloop


async def trace_all_equilibria( # pylint: disable=too-many-locals
        agame0, agame1, *, regret_thresh=1e-3, dist_thresh=0.1, max_step=0.1,
        executor=None, **innerloop_args):
    """Trace out all equilibria between all games

    Parameters
    ----------
    agame0 : AsyncGame
        The game that is played when time is 0.
    agame1 : AsyncGame
        The game that is played when time is 1.
    regret_thresh : float, optional
        The threshold for epsilon regret for equilibria returned.
    exectutor : Executor, optional
        The executor to run computation intensive operations in.
    """
    utils.check(
        rsgame.empty_copy(agame0) == rsgame.empty_copy(agame1),
        'games must have same structure')
    loop = asyncio.get_event_loop()
    trace_args = dict(regret_thresh=regret_thresh, max_step=max_step)
    innerloop_args.update(
        executor=executor, regret_thresh=regret_thresh,
        dist_thresh=dist_thresh)

    async def trace_eqm(eqm, prob):
        """Trace and equilibrium out from prob"""
        game0 = agame0.get_game()
        game1 = agame1.get_game()
        (pr0, eqa0), (pr1, eqa1) = await asyncio.gather(
            loop.run_in_executor(executor, functools.partial(
                trace.trace_equilibrium, game0, game1, prob, eqm, 0,
                **trace_args)),
            loop.run_in_executor(executor, functools.partial(
                trace.trace_equilibrium, game0, game1, prob, eqm, 1,
                **trace_args)))
        return (
            np.concatenate([pr0[::-1], pr1[1:]]),
            np.concatenate([eqa0[::-1], eqa1[1:]]))

    async def trace_between(lower, upper):
        """Trace between times lower and upper"""
        if upper <= lower:
            return ()
        mid = (lower + upper) / 2

        midgame = asyncgame.mix(agame0, agame1, mid)
        eqa = await innerloop.inner_loop(midgame, **innerloop_args)

        if not eqa.size:  # pragma: no cover
            logging.warning('found no equilibria in %s', midgame)
            return ()

        traces = await asyncio.gather(*[
            trace_eqm(eqm, mid) for eqm in eqa])

        lupper = min(t[0] for t, _ in traces)
        ulower = max(t[-1] for t, _ in traces)
        logging.warning(
            'traced %s out to %g - %g', midgame, lupper, ulower)

        lower_traces, upper_traces = await asyncio.gather(
            trace_between(lower, lupper), trace_between(ulower, upper))
        # Lazily extend them
        return itertools.chain(lower_traces, traces, upper_traces)

    traces = list(await trace_between(0.0, 1.0))
    game0, game1 = agame0.get_game(), agame1.get_game()
    traces = _merge_traces(game0, game1, traces, dist_thresh, trace_args)
    for probs, eqa in traces:
        _smooth_trace(game0, game1, probs, eqa, trace_args)
        _smooth_trace(game0, game1, probs[::-1], eqa[::-1], trace_args)
    return sorted(traces, key=lambda tr: (tr[0][0], tr[0][-1]))


def _trace_distance(game0, game1, trace1, trace2, trace_args):
    """Compute the distance between traces

    This uses interpolation to estimate each trace at arbitrary points in time
    and then computes the average time-weighted norm between the traces."""
    time1, eqa1 = trace1
    time2, eqa2 = trace2
    if time1[-1] < time2[0] or time2[-1] < time1[0]:
        return np.inf

    tmin = max(time1[0], time2[0])
    tmax = min(time1[-1], time2[-1])

    # XXX This sorted merge could be more efficient
    times = np.concatenate([
        time1[(tmin <= time1) & (time1 <= tmax)],
        time2[(tmin <= time2) & (time2 <= tmax)]])
    times.sort()

    eqa1i = trace.trace_interpolate(
        game0, game1, time1, eqa1, times, **trace_args)
    eqa2i = trace.trace_interpolate(
        game0, game1, time2, eqa2, times, **trace_args)
    errs = np.linalg.norm(eqa1i - eqa2i, axis=1)
    return np.diff(times).dot(errs[:-1] + errs[1:]) / 2 / (tmax - tmin)


def _merge_traces(game0, game1, traces, thresh, trace_args): # pylint: disable=too-many-locals
    """Merge a list of traces

    Parameters
    ----------
    traces : [(ts, eqa)]
        A list of traces, which are themselves tuples of times and equilibria.
    thresh : float
        How similar traces need to be in order to be merged. This is the
        average norm between the traces.
    interp : str
        The way to interpolate between equilibria. This is passed to interp1d.
    """
    distances = np.zeros([len(traces)] * 2)
    for i, trace1 in enumerate(traces):
        for j, trace2 in enumerate(traces[:i]):
            distances[i, j] = distances[j, i] = _trace_distance(
                game0, game1, trace1, trace2, trace_args)
    num, comps = csgraph.connected_components(distances <= thresh, False)
    new_traces = []
    for i in range(num):
        times = np.concatenate([
            t for (t, _), m in zip(traces, comps == i) if m])
        eqa = np.concatenate([
            eqms for (_, eqms), m in zip(traces, comps == i) if m])
        inds = np.argsort(times)
        new_traces.append((times[inds], eqa[inds]))
    return new_traces


def _smooth_trace(game0, game1, probs, eqa, trace_args):
    """Smooth the equilibria in a trace in place

    Smoothing attempts to trace out from one time to an adjacent time. If the
    new point has lower regret, it's taken instead. This onle goes one
    direction, so it should be repeated for reversed views.
    """
    for (pfrom, pto), (eqmfrom, eqmto) in zip(
            utils.subsequences(probs), utils.subsequences(eqa)):
        (*_, pres), (*_, eqmres) = trace.trace_equilibrium( # pylint: disable=too-many-star-expressions
            game0, game1, pfrom, eqmfrom, pto, **trace_args)
        if np.isclose(pres, pto):
            mixgame = rsgame.mix(game0, game1, pto)
            regto = regret.mixture_regret(mixgame, eqmto)
            regres = regret.mixture_regret(mixgame, eqmres)
            if regres < regto:
                np.copyto(eqmto, eqmres)



"""Test zip scheduler"""
import asyncio
import io
import json
import zipfile

import numpy as np
import pytest
from gameanalysis import rsgame

from egta import countsched
from egta import zipsched


def batch_to_zip(bash):
    """Convert the string of a bash script into a zip file"""
    buff = io.BytesIO()
    with zipfile.ZipFile(buff, 'w') as zfil:
        zfil.writestr('package/defaults.json', json.dumps({}))
        zfil.writestr('package/script/batch', '#!/usr/bin/env bash\n' + bash)
    return buff


@pytest.fixture(name='jgame')
def fix_jgame():
    """Fixture for the common game json"""
    with open('cdasim/game.json') as fil:
        return json.load(fil)


@pytest.fixture(name='conf')
def fix_conf(jgame):
    """Fixture for a default configuration"""
    return jgame['configuration']


@pytest.fixture(name='game')
def fix_game(jgame):
    """Fixture a game"""
    return rsgame.empty_json(jgame)


@pytest.mark.asyncio
async def test_basic_profile(conf, game):
    """Test scheduling a basic profile"""
    profs = game.random_profiles(20)
    zipf = 'cdasim/cdasim.zip'

    with zipsched.zipsched(game, conf, zipf) as sched:
        assert str(sched) is not None
        assert rsgame.empty_copy(sched) == game
        awaited = await asyncio.gather(*[
            sched.sample_payoffs(p) for p in profs])

    pays = np.stack(awaited)
    assert pays.shape == profs.shape
    assert np.allclose(pays[profs == 0], 0)
    assert np.any(pays != 0)
    # XXX This line exists to fool duplication check 1


@pytest.mark.asyncio
async def test_no_conf(game):
    """Test scheduler without a configuration"""
    profs = game.random_profiles(20)
    zipf = 'cdasim/cdasim.zip'

    with zipsched.zipsched(game, {}, zipf) as sched:
        assert rsgame.empty_copy(sched) == game
        awaited = await asyncio.gather(*[
            sched.sample_payoffs(p) for p in profs])

    pays = np.stack(awaited)
    assert pays.shape == profs.shape
    assert np.allclose(pays[profs == 0], 0)
    assert np.any(pays != 0)
    # XXX This line exists to fool duplication check 2


@pytest.mark.asyncio
@pytest.mark.parametrize('count', [2, 3])
async def test_simultaneous_obs(conf, game, count):
    """Test scheduler appropriately schedules simultaneous observations"""
    profs = game.random_profiles(10).repeat(2, 0)
    zipf = 'cdasim/cdasim.zip'

    with zipsched.zipsched(
        game, conf, zipf, simultaneous_obs=count) as sched:
        sched = countsched.countsched(sched, count)
        awaited = await asyncio.gather(*[
            sched.sample_payoffs(p) for p in profs])

    pays = np.stack(awaited)
    assert pays.shape == profs.shape
    assert np.allclose(pays[profs == 0], 0)
    assert np.any(pays != 0)
    # XXX This line exists to fool duplication check 3


@pytest.mark.asyncio
async def test_get_fail(conf, game):
    """Test zip file that fails"""
    prof = game.random_profile()
    zipf = batch_to_zip('sleep 0.2 && false')

    with zipsched.zipsched(game, conf, zipf) as sched:
        future = sched.sample_payoffs(prof)
        with pytest.raises(ValueError):
            await future


@pytest.mark.asyncio
async def test_bad_zipfile(conf, game):
    """Test that improper zip file fails"""
    zipf = 'nonexistent'
    with pytest.raises(FileNotFoundError):
        with zipsched.zipsched(game, conf, zipf):
            pass  # pragma: no cover


@pytest.mark.asyncio
async def test_no_obs(conf, game):
    """Test zip sched that doesn't write observations"""
    prof = game.random_profile()
    zipf = batch_to_zip('sleep 0.2')

    with zipsched.zipsched(game, conf, zipf) as sched:
        future = sched.sample_payoffs(prof)
        with pytest.raises(ValueError):
            await future



"""Tests for cli"""
import contextlib
import io
import itertools
import json
import random
import subprocess
import sys
import traceback
from os import path
from unittest import mock

import numpy as np
import pytest
from egtaonline import api
from egtaonline import mockserver
from gameanalysis import gamegen
from gameanalysis import gamereader
from gameanalysis import rsgame
from gameanalysis import utils

from egta import __main__ as main


BASE = path.dirname(path.realpath(__file__))
EGTA = path.join(BASE, '..', 'bin', 'egta')
SIM_DIR = path.join(BASE, '..', 'cdasim')
SIM_GAME_FILE = path.join(SIM_DIR, 'small_game.json')


async def run(*args):
    """Run a command line and return if it ran successfully"""
    try:
        await main.amain(*args)
    except SystemExit as ex:
        return not int(str(ex))
    except Exception: # pylint: disable=broad-except
        traceback.print_exc()
        return False
    return True


def stdin(inp):
    """Patch stdin with input"""
    return mock.patch.object(sys, 'stdin', io.StringIO(inp))


def stdout():
    """Patch stdout and return stringio"""
    return contextlib.redirect_stdout(io.StringIO())


def stderr():
    """Patch stderr and return stringio"""
    return contextlib.redirect_stderr(io.StringIO())


@pytest.fixture(scope='session', name='jgame')
def fix_jgame():
    """Return json version of sim game"""
    with open(SIM_GAME_FILE) as fil:
        return json.load(fil)


@pytest.fixture(scope='session', name='game')
def fix_game(jgame):
    """Return game"""
    return gamegen.gen_profiles(gamereader.loadj(jgame))


@pytest.fixture(scope='session', name='game_sched')
def fix_game_sched(game, tmpdir_factory):
    """Create game scheduler string"""
    game_file = str(tmpdir_factory.mktemp('games').join('game.json'))
    with open(game_file, 'w') as fil:
        json.dump(game.to_json(), fil)
    return 'game:game:{}'.format(game_file)


@pytest.fixture(scope='session', name='sim_sched')
def fix_sim_sched(jgame, tmpdir_factory):
    """Create sim scheduler string"""
    conf_file = str(tmpdir_factory.mktemp('conf').join('conf.json'))
    with open(conf_file, 'w') as fil:
        json.dump(jgame['configuration'], fil)
    sim = ' '.join([
        path.join(BASE, '..', 'bin', 'python'), path.join(SIM_DIR, 'sim.py'),
        '1', '--single'])
    return 'sim:game:{},conf:{},command:{}'.format(
        SIM_GAME_FILE, conf_file, sim)


@pytest.fixture(scope='session', name='zip_sched')
def fix_zip_sched():
    """Create zip scheduler string"""
    zip_file = path.join(SIM_DIR, 'cdasim.zip')
    return 'zip:game:{},zipf:{}'.format(SIM_GAME_FILE, zip_file)


def assert_term(sleep1, sleep2, *args):
    """Assert that command handles sigterm"""
    proc = subprocess.Popen((EGTA,) + args)
    try:
        proc.wait(sleep1)
        # proc.wait should throw exception and pass i.e. process is still
        # running after sleep1
        proc.kill()  # pragma: no cover
        assert False, 'process finished during first sleep'  # pragma: no cover
    except subprocess.TimeoutExpired:
        pass
    proc.terminate()
    try:
        assert proc.wait(sleep2), 'process succeeded after termination'
    except subprocess.TimeoutExpired:  # pragma: no cover
        # proc.terminate should terminate process before sleep2 is reached, so
        # we should never trigger except
        proc.kill()
        assert False, "process didn't terminate in second sleep"


@pytest.mark.asyncio
async def test_help():
    """Test help works"""
    assert not await run()
    assert not await run('--fail')
    with stderr() as err:
        assert await run('--help'), err.getvalue()
    with stderr() as err:
        assert await run('brute', '--help'), err.getvalue()
    with stderr() as err:
        assert await run('quiesce', '--help'), err.getvalue()
    with stderr() as err:
        assert await run('boot', '--help'), err.getvalue()
    with stderr() as err:
        assert await run('trace', '--help'), err.getvalue()


@pytest.mark.asyncio
@pytest.mark.parametrize('red', [
    [],
    ['--dpr', 'buyers:2,sellers:2'],
    ['--hr', 'buyers:2,sellers:2'],
])
async def test_brute_game(game, game_sched, red):
    """Test brute game"""
    with stdout() as out, stderr() as err:
        assert await run(
            'brute', game_sched + ',count:2', *red), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_brute_game_tag(game, game_sched):
    """Test brute game with tag"""
    with stdout() as out, stderr() as err:
        assert await run(
            '--tag', 'test', 'brute', game_sched + ',count:2'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_brute_game_restriction(game, game_sched):
    """Test brute game with restriction"""
    rest = json.dumps(game.restriction_to_json(game.random_restriction()))
    with stdin(rest), stdout() as out, stderr() as err:
        assert await run(
            'brute', game_sched + ',count:2', '--restrict', '-'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


def test_brute_game_term(game_sched):
    """Test brute game handles sigterm"""
    assert_term(0.5, 0.5, 'brute', game_sched + ',count:2')


@pytest.mark.asyncio
async def test_prof_data(game, game_sched, tmpdir):
    """Test game scheduler"""
    prof_file = str(tmpdir.join('profs.json'))
    with stdout() as out, stderr() as err:
        assert await run(
            'brute', game_sched + ',sample:,save:' + prof_file), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])
    with open(prof_file) as fil:
        prof_game = gamereader.load(fil)
    assert rsgame.empty_copy(game) == rsgame.empty_copy(prof_game)


@pytest.mark.asyncio
async def test_sim(game, sim_sched):
    """Test sim sched"""
    with stdout() as out, stderr() as err:
        assert await run(
            'brute', sim_sched, '--dpr', 'buyers:2;sellers:2',
            '--min-reg'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_sim_in(game, sim_sched):
    """Test sim with stdin"""
    with stdin(json.dumps(game.to_json())), stdout() as out, stderr() as err:
        assert await run(
            'brute', sim_sched + ',game:-', '--dpr', 'buyers:2;sellers:2',
            '--min-reg'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_zip(game, zip_sched):
    """Test zip scheduler"""
    with stdout() as out, stderr() as err:
        assert await run(
            'brute', zip_sched, '--dpr', 'buyers:2;sellers:2',
            '--min-reg'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_zip_in(game, zip_sched):
    """Test zip with stdin"""
    with stdin(json.dumps(game.to_json())), stdout() as out, stderr() as err:
        assert await run(
            'brute', zip_sched + ',game:-', '--dpr', 'buyers:2;sellers:2',
            '--min-reg'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_zip_conf_in(game, zip_sched):
    """Test zip with configuration in stdin"""
    conf = json.dumps({'max_value': 2})
    with stdin(conf), stdout() as out, stderr() as err:
        assert await run(
            'brute', zip_sched + ',conf:-', '--dpr', 'buyers:2;sellers:2',
            '--min-reg'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_zip_conf_file(game, zip_sched, tmpdir):
    """Test zipsched with conf file"""
    conf_file = str(tmpdir.join('conf.json'))
    with open(conf_file, 'w') as fil:
        json.dump({'max_value': 2}, fil)
    with stdout() as out, stderr() as err:
        assert await run(
            'brute', zip_sched + ',conf:' + conf_file, '--dpr',
            'buyers:2;sellers:2', '--min-reg'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


def test_brute_sim_term(sim_sched):
    """Test brute sim sigterm"""
    assert_term(0.5, 0.5, 'brute', sim_sched)


@pytest.mark.asyncio
async def test_sim_delayed_fail(tmpdir):
    """Test delayed simulation fail"""
    game = rsgame.empty(4, 3)
    game_file = str(tmpdir.join('game.json'))
    with open(game_file, 'w') as fil:
        json.dump(game.to_json(), fil)
    script = str(tmpdir.join('script.sh'))
    with open(script, 'w') as fil:
        fil.write('sleep 1 && false')
    sched = 'sim:game:{},command:bash {}'.format(game_file, script)
    with stdin(json.dumps({})):
        assert not await run(
            'brute', sched)


@pytest.mark.asyncio
async def test_sim_conf(game, sim_sched, tmpdir):
    """Test sim with configuration"""
    conf_file = str(tmpdir.join('conf.json'))
    with open(conf_file, 'w') as fil:
        json.dump({
            'markup': 'standard',
            'max_value': 1,
            'arrivals': 'simple',
            'market': 'call',
        }, fil)
    with stdout() as out, stderr() as err:
        assert await run(
            'brute', '--dpr', 'buyers:2;sellers:2', sim_sched + ',conf:' +
            conf_file), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_sim_conf_in(game, sim_sched):
    """Test sim with configuration in stdin"""
    conf = json.dumps({
        'markup': 'standard',
        'max_value': 1,
        'arrivals': 'simple',
        'market': 'call',
    })
    with stdin(conf), stdout() as out, stderr() as err:
        assert await run(
            'brute', '--dpr', 'buyers:2;sellers:2',
            sim_sched + ',conf:-'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_innerloop(game, game_sched):
    """Test basic inner loop"""
    with stdout() as out, stderr() as err:
        assert await run('quiesce', game_sched), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


def test_innerloop_game_term(game_sched):
    """Test that inner loop game handles sigterm"""
    assert_term(0.5, 0.5, 'quiesce', game_sched)


def test_innerloop_sim_term(sim_sched):
    """Test that inner loop sim handles sigterm"""
    assert_term(0.5, 0.5, 'quiesce', sim_sched)


@pytest.mark.asyncio
@pytest.mark.parametrize('red', [
    [],
    ['--dpr', 'buyers:2,sellers:2'],
    ['--hr', 'buyers:2,sellers:2'],
])
async def test_innerloop_red(game, game_sched, red):
    """Test inner loop with reductions"""
    with stdout() as out, stderr() as err:
        assert await run(
            'quiesce', game_sched, *red), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


async def get_eosched(server, game, name):
    """Get an eo sched"""
    async with api.api() as egta:
        sim = await egta.get_simulator(server.create_simulator(
            name, '1', delay_dist=lambda: random.random() / 100))
        await sim.add_strategies(dict(zip(game.role_names, game.strat_names)))
        eogame = await sim.create_game(name, game.num_players)
        await eogame.add_symgroups(list(zip(
            game.role_names, game.num_role_players, game.strat_names)))
        return 'eo:game:{:d},mem:2048,time:60,sleep:0.1'.format(
            eogame['id'])


@pytest.mark.asyncio
async def test_brute_egta_game(game):
    """Test brute game with egta sched"""
    async with mockserver.server() as server:
        sched = await get_eosched(server, game, 'game')
        with stdout() as out, stderr() as err:
            assert await run(
                'brute', sched, '--dpr', 'buyers:2;sellers:2',
                '--min-reg'), err.getvalue()
            for eqm in json.loads(out.getvalue()):
                game.mixture_from_json(eqm['equilibrium'])


def verify_trace_json(game, traces):
    """Verify traced json"""
    for trace in traces:
        for point in trace:
            assert point.keys() == {'regret', 'equilibrium', 't'}
            game.mixture_from_json(point['equilibrium'])
        assert utils.is_sorted(trace, key=lambda t: t['t'])


@pytest.mark.asyncio
@pytest.mark.parametrize('red', [
    [],
    ['--dpr', 'buyers:2,sellers:2'],
    ['--hr', 'buyers:2,sellers:2'],
])
async def test_trace(game, sim_sched, tmpdir, red):
    """Test tracing"""
    conf_file = str(tmpdir.join('conf.json'))
    with open(conf_file, 'w') as fil:
        json.dump({
            'markup': 'standard',
            'max_value': 1,
            'arrivals': 'simple',
            'market': 'call',
        }, fil)
    prof_data = str(tmpdir.join('data.json'))
    with stdout() as out, stderr() as err:
        assert await run(
            'trace', sim_sched + ',save:' + prof_data, sim_sched + ',conf:' +
            conf_file, *red), err.getvalue()
    traces = json.loads(out.getvalue())
    verify_trace_json(game, traces)
    with open(prof_data) as fil:
        data = gamereader.load(fil)
    assert rsgame.empty_copy(game) == rsgame.empty_copy(data)


def add_singleton_role(game, index, role_name, strat_name, num_players):
    """Add a singleton role to a game"""
    role_names = list(game.role_names)
    role_names.insert(index, role_name)
    strat_names = list(game.strat_names)
    strat_names.insert(index, [strat_name])
    role_players = np.insert(game.num_role_players, index, num_players)
    return rsgame.empty_names(role_names, role_players, strat_names)


@pytest.mark.asyncio
async def test_trace_norm():
    """Test tracing"""
    base_game = rsgame.empty([3, 2], [2, 3])
    async with mockserver.server() as server:
        game0 = add_singleton_role(base_game, 0, 'a', 'sx', 1)
        sched0 = await get_eosched(server, game0, 'game0')
        game1 = add_singleton_role(base_game, 1, 'r00', 's0', 3)
        sched1 = await get_eosched(server, game1, 'game1')
        with stdout() as out, stderr() as err:
            assert await run('trace', sched0, sched1), err.getvalue()
        traces = json.loads(out.getvalue())
    verify_trace_json(base_game, traces)


@pytest.mark.asyncio
async def test_boot_game(game, game_sched, tmpdir):
    """Test bootstrapping a game"""
    mix_file = str(tmpdir.join('mix.json'))
    with open(mix_file, 'w') as fil:
        json.dump(game.mixture_to_json(game.random_mixture()), fil)
    with stdout() as out, stderr() as err:
        assert await run(
            'boot', game_sched, mix_file, '10'), err.getvalue()
        results = json.loads(out.getvalue())
    assert {'total', 'buyers', 'sellers'} == results.keys()
    for val in results.values():
        assert {'surplus', 'regret', 'response'} == val.keys()


@pytest.mark.asyncio
async def test_boot_symmetric(tmpdir):
    """Test bootstrapping a symmetric game"""
    game = gamegen.game(4, 3)
    mix_file = str(tmpdir.join('mix.json'))
    with open(mix_file, 'w') as fil:
        json.dump(game.mixture_to_json(game.random_mixture()), fil)
    with stdin(json.dumps(game.to_json())), stdout() as out, stderr() as err:
        assert await run(
            'boot', 'game', mix_file, '10'), err.getvalue()
        results = json.loads(out.getvalue())
    assert {'surplus', 'regret', 'response'} == results.keys()


@pytest.mark.asyncio
async def test_boot_symmetric_percs(tmpdir):
    """Test bootstrapping a symmetric game with percentiles"""
    game = gamegen.game(4, 3)
    mix_file = str(tmpdir.join('mix.json'))
    with open(mix_file, 'w') as fil:
        json.dump(game.mixture_to_json(game.random_mixture()), fil)
    with stdin(json.dumps(game.to_json())), stdout() as out, stderr() as err:
        assert await run(
            'boot', 'game', mix_file, '10', '-p95'), err.getvalue()
        results = json.loads(out.getvalue())
    assert {'95', 'mean'} == results.keys()
    assert {'surplus', 'regret', 'response'} == results['mean'].keys()
    assert {'surplus', 'regret'} == results['95'].keys()


@pytest.mark.asyncio
async def test_boot_game_percs(game, game_sched):
    """Test bootstrapping a game with percentiles"""
    mix = json.dumps(game.mixture_to_json(game.random_mixture()))
    with stdin(mix), stdout() as out, stderr() as err:
        assert await run(
            'boot', game_sched, '-', '20', '--percentile', '95',
            '-p99'), err.getvalue()
        results = json.loads(out.getvalue())
    assert {'mean', '99', '95'} == results.keys()
    for val in results.values():
        assert {'total', 'buyers', 'sellers'} == val.keys()
    for val in results['mean'].values():
        assert {'surplus', 'regret', 'response'} == val.keys()
    for val in itertools.chain(results['95'].values(), results['99'].values()):
        assert {'surplus', 'regret'} == val.keys()


@pytest.mark.asyncio
async def test_boot_sim(game, sim_sched):
    """Test bootstrapping a sim scheduler"""
    mix = json.dumps(game.mixture_to_json(game.random_mixture()))
    with stdin(mix), stdout() as out, stderr() as err:
        assert await run(
            'boot', sim_sched, '-', '50', '--chunk-size', '10'), err.getvalue()
        results = json.loads(out.getvalue())
    assert {'total', 'buyers', 'sellers'} == results.keys()
    for val in results.values():
        assert {'surplus', 'regret', 'response'} == val.keys()



"""module for games that require asynchronous profile access"""
import abc
import asyncio

import numpy as np
from gameanalysis import rsgame
from gameanalysis import utils


class _AsyncGame(rsgame._GameLike): # pylint: disable=protected-access
    """An asynchronous game

    Supports asynchronous methods for ensuring particular payoff data"""

    @abc.abstractmethod
    def get_game(self):
        """Return all data available"""
        pass  # pragma: no cover

    @abc.abstractmethod
    async def get_restricted_game(self, rest):
        """Return a complete restricted game"""
        pass  # pragma: no cover

    @abc.abstractmethod
    async def get_deviation_game(self, rest, role_index=None):
        """Return a game with payoff data for all deviations

        If role index is specified, only deviations from that role are
        necessary."""
        pass  # pragma: no cover


class _CompleteAsyncGame(_AsyncGame):
    """A wrapper for a complete RsGame"""
    def __init__(self, game):
        super().__init__(
            game.role_names, game.strat_names, game.num_role_players)
        self._game = game

    def get_game(self):
        return self._game

    async def get_restricted_game(self, rest):
        return self._game.restrict(rest)

    async def get_deviation_game(self, rest, role_index=None):
        return self._game

    def __eq__(self, other):
        return super().__eq__(other) and self._game == other._game # pylint: disable=protected-access

    def __hash__(self):
        return hash(self._game)

    def __str__(self):
        return repr(self._game)


def wrap(game):
    """Wrap a CompleteGame as an AsyncGame"""
    utils.check(game.is_complete(), 'must use a complete game')
    return _CompleteAsyncGame(game)


class _MixedAsyncGame(_AsyncGame):
    """A lazy merging of two async games"""
    def __init__(self, agame0, agame1, prob):
        super().__init__(
            agame0.role_names, agame0.strat_names, agame0.num_role_players)
        self._agame0 = agame0
        self._agame1 = agame1
        self._prob = prob

    def get_game(self):
        return rsgame.mix(
            self._agame0.get_game(), self._agame1.get_game(), self._prob)

    async def get_restricted_game(self, rest):
        game0, game1 = await asyncio.gather(
            self._agame0.get_restricted_game(rest),
            self._agame1.get_restricted_game(rest))
        return rsgame.mix(game0, game1, self._prob)

    async def get_deviation_game(self, rest, role_index=None):
        game0, game1 = await asyncio.gather(
            self._agame0.get_deviation_game(rest, role_index),
            self._agame1.get_deviation_game(rest, role_index))
        return rsgame.mix(game0, game1, self._prob)

    def __eq__(self, othr):
        # pylint: disable-msg=protected-access
        return (
            super().__eq__(othr) and ((
                self._agame0 == othr._agame0 and
                self._agame1 == othr._agame1 and
                np.isclose(self._prob, othr._prob)
            ) or (
                self._agame0 == othr._agame1 and
                self._agame1 == othr._agame0 and
                np.isclose(self._prob, 1 - othr._prob)
            )))

    @utils.memoize
    def __hash__(self):
        return hash(frozenset([self._agame0, self._agame1]))

    def __str__(self):
        return '{} - {:g} - {}'.format(
            self._agame0, self._prob, self._agame1)


def mix(agame0, agame1, prob):
    """Mix two async games"""
    utils.check(
        rsgame.empty_copy(agame0) == rsgame.empty_copy(agame1),
        'games must have identically structure')
    return _MixedAsyncGame(agame0, agame1, prob)



"""An abstract profile scheduler"""
import abc

from gameanalysis import rsgame


class _Scheduler(rsgame._GameLike): # pylint: disable=protected-access
    """A profile scheduler

    It must be a context manager for its resources."""

    @abc.abstractmethod
    async def sample_payoffs(self, profile):
        """Schedule a profile

        Return a future for the payoff data"""
        pass  # pragma: no cover


class _OpenableScheduler(_Scheduler):
    """Scheduler that is opened"""
    @abc.abstractmethod
    def open(self):
        """Open the scheduler"""
        pass # pragma: no cover

    @abc.abstractmethod
    def close(self):
        """Close the scheduler"""
        pass # pragma: no cover

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args):
        self.close()


class _AOpenableScheduler(_Scheduler):
    """Scheduler that is asynchronously opened"""
    @abc.abstractmethod
    async def aopen(self):
        """Open the scheduler"""
        pass # pragma: no cover

    @abc.abstractmethod
    async def aclose(self):
        """Close the scheduler"""
        pass # pragma: no cover

    async def __aenter__(self):
        await self.aopen()
        return self

    async def __aexit__(self, *args):
        await self.aclose()



"""Command line module for game scheduler creation"""
import sys

from gameanalysis import paygame
from gameanalysis import gamereader

from egta import gamesched


async def create_scheduler(
        game: 'A file with the game data. `-` is interpreted as stdin.' = '-',
        sample: """Treat the game as a sample game. This doesn't require a
        value, just a trialing colon.""" = None, **_):
    """A scheduler that samples payoff data from an existing game. If sample is
    specified, payoffs will be a random payoff from each payoff for the
    profile."""
    if game == '-':
        rsgame = gamereader.load(sys.stdin)
    else:
        with open(game) as fil:
            rsgame = gamereader.load(fil)

    if sample is not None: # pylint: disable=no-else-return
        return Sgs(paygame.samplegame_copy(rsgame))
    else:
        return Rgs(rsgame)


class Sgs(gamesched._SampleGameScheduler): # pylint: disable=protected-access
    """Async context manager SampleGameScheduler"""
    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        pass


class Rgs(gamesched._RsGameScheduler): # pylint: disable=protected-access
    """Async context manager RsGameScheduler"""
    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        pass
